import random

from django.conf import settings


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    return available_metrics


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-15 14:16
from __future__ import unicode_literals

import uuid

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import scheduler.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("djcelery", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("frequency", models.IntegerField(blank=True, null=True)),
                ("triggered", models.IntegerField(default=0)),
                (
                    "cron_definition",
                    models.CharField(
                        max_length=500,
                        null=True,
                        validators=[scheduler.models.validate_crontab],
                    ),
                ),
                (
                    "interval_definition",
                    models.CharField(
                        max_length=100,
                        null=True,
                        validators=[scheduler.models.validate_interval],
                    ),
                ),
                ("endpoint", models.CharField(max_length=500)),
                ("auth_token", models.CharField(blank=True, max_length=500, null=True)),
                ("payload", django.contrib.postgres.fields.jsonb.JSONField(default={})),
                ("next_send_at", models.DateTimeField(blank=True, null=True)),
                ("enabled", models.BooleanField(default=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "celery_cron_definition",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djcelery.CrontabSchedule",
                    ),
                ),
                (
                    "celery_interval_definition",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djcelery.IntervalSchedule",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="schedules_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="schedules_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        )
    ]



import json
from uuid import uuid4

import requests
from celery.task import Task
from celery.utils.log import get_task_logger
from django.conf import settings
from django.utils.timezone import now
from djcelery.models import CrontabSchedule, IntervalSchedule
from requests import exceptions as requests_exceptions
from seed_services_client.metrics import MetricsApiClient

from seed_scheduler import utils

from .models import QueueTaskRun, Schedule, ScheduleFailure

logger = get_task_logger(__name__)


class DeliverHook(Task):
    def run(self, target, payload, instance_id=None, hook_id=None, **kwargs):
        """
        target:     the url to receive the payload.
        payload:    a python primitive data structure
        instance_id:   a possibly None "trigger" instance ID
        hook_id:       the ID of defining Hook object
        """
        requests.post(
            url=target,
            data=json.dumps(payload),
            headers={
                "Content-Type": "application/json",
                "Authorization": "Token %s" % settings.HOOK_AUTH_TOKEN,
            },
        )


def deliver_hook_wrapper(target, payload, instance, hook):
    if instance is not None:
        instance_id = instance.id
    else:
        instance_id = None
    kwargs = dict(
        target=target, payload=payload, instance_id=instance_id, hook_id=hook.id
    )
    DeliverHook.apply_async(kwargs=kwargs)


class DeliverTask(Task):

    """
    Task to deliver scheduled hook
    """

    name = "seed_scheduler.scheduler.tasks.deliver_task"
    default_retry_delay = 5
    max_retries = 5

    def run(self, schedule_id, auth_token, endpoint, payload, **kwargs):
        """
        Runs an instance of a scheduled task
        """
        log = self.get_logger(**kwargs)
        log.info("Running instance of <%s>" % (schedule_id,))
        if self.request.retries > 0:
            retry_delay = utils.calculate_retry_delay(self.request.retries)
        else:
            retry_delay = self.default_retry_delay

        headers = {"Content-Type": "application/json"}
        if auth_token is not None:
            headers["Authorization"] = "Token %s" % auth_token
        try:
            response = requests.post(
                url=endpoint,
                data=json.dumps(payload),
                headers=headers,
                timeout=settings.DEFAULT_REQUEST_TIMEOUT,
            )
            # Expecting a 201, raise for errors.
            response.raise_for_status()

            Schedule.objects.filter(id=schedule_id).update(last_run=now())
        except requests_exceptions.ConnectionError as exc:
            log.info("Connection Error to endpoint: %s" % endpoint)
            fire_metric.delay("scheduler.deliver_task.connection_error.sum", 1)
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.HTTPError as exc:
            # Recoverable HTTP errors: 500, 401
            log.info("Request failed due to status: %s" % exc.response.status_code)
            metric_name = (
                "scheduler.deliver_task.http_error.%s.sum" % exc.response.status_code
            )
            fire_metric.delay(metric_name, 1)
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.Timeout as exc:
            log.info("Request failed due to timeout")
            fire_metric.delay("scheduler.deliver_task.timeout.sum", 1)
            self.retry(exc=exc, countdown=retry_delay)

        return True

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        if self.request.retries == self.max_retries:
            if "schedule_id" in kwargs:
                schedule_id = kwargs["schedule_id"]
            else:
                schedule_id = args[0]
            ScheduleFailure.objects.create(
                schedule_id=schedule_id,
                initiated_at=self.request.eta,
                reason=einfo.exception.message,
                task_id=task_id,
            )
        super(DeliverTask, self).on_failure(exc, task_id, args, kwargs, einfo)


deliver_task = DeliverTask()


class QueueTasks(Task):

    """
    Task to queue delivery of scheduled hooks
    """

    name = "seed_scheduler.scheduler.tasks.queue_tasks"
    ignore_result = True

    def run(self, schedule_type, lookup_id, **kwargs):
        """
        Loads Schedule linked to provided lookup
        """
        log = self.get_logger(**kwargs)
        log.info("Queuing <%s> <%s>" % (schedule_type, lookup_id))

        task_run = QueueTaskRun()
        task_run.task_id = self.request.id or uuid4()
        task_run.started_at = now()
        tr_qs = QueueTaskRun.objects

        # Load the schedule active items
        schedules = Schedule.objects.filter(enabled=True)
        if schedule_type == "crontab":
            schedules = schedules.filter(celery_cron_definition=lookup_id)
            tr_qs = tr_qs.filter(celery_cron_definition=lookup_id)
            scheduler_type = CrontabSchedule
            task_run.celery_cron_definition_id = lookup_id
        elif schedule_type == "interval":
            schedules = schedules.filter(celery_interval_definition=lookup_id)
            tr_qs = tr_qs.filter(celery_interval_definition=lookup_id)
            scheduler_type = IntervalSchedule
            task_run.celery_interval_definition_id = lookup_id

        # Confirm that this task should run now based on last run time.
        try:
            last_task_run = tr_qs.latest("started_at")
        except QueueTaskRun.DoesNotExist:
            # No previous run so it is safe to continue.
            pass
        else:
            # This basicly replicates what celery beat is meant to do, but
            # we can't trust celery beat and django-celery to always accurately
            # update their own last run time.
            sched = scheduler_type.objects.get(id=lookup_id)
            due, due_next = sched.schedule.is_due(last_task_run.started_at)
            if not due and due_next >= settings.DEFAULT_CLOCK_SKEW_SECONDS:
                return (
                    "Aborted Queuing <%s> <%s> due to last task run (%s) "
                    "at %s"
                    % (
                        schedule_type,
                        lookup_id,
                        last_task_run.id,
                        last_task_run.started_at,
                    )
                )

        task_run.save()
        # create tasks for each active schedule
        queued = 0
        schedules = schedules.values("id", "auth_token", "endpoint", "payload")
        for schedule in schedules.iterator():
            schedule["schedule_id"] = str(schedule.pop("id"))
            DeliverTask.apply_async(kwargs=schedule)
            queued += 1

        task_run.completed_at = now()
        task_run.save()
        return "Queued <%s> Tasks" % (queued,)


queue_tasks = QueueTasks()


def get_metric_client(session=None):
    return MetricsApiClient(
        url=settings.METRICS_URL, auth=settings.METRICS_AUTH, session=session
    )


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """

    name = "seed_scheduler.scheduler.tasks.fire_metric"

    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {metric_name: metric_value}
        metric_client = get_metric_client(session=session)
        metric_client.fire_metrics(**metric)
        return "Fired metric <%s> with value <%s>" % (metric_name, metric_value)


fire_metric = FireMetric()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed schedules.
    """

    name = "seed_scheduler.scheduler.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        """
        Runs an instance of a scheduled task
        """
        log = self.get_logger(**kwargs)
        failures = ScheduleFailure.objects
        log.info("Attempting to requeue <%s> failed schedules" % failures.count())
        for failure in failures.iterator():
            schedule = Schedule.objects.values(
                "id", "auth_token", "endpoint", "payload"
            )
            schedule = schedule.get(id=failure.schedule_id)
            schedule["schedule_id"] = str(schedule.pop("id"))
            # Cleanup the failure before requeueing it.
            failure.delete()
            DeliverTask.apply_async(kwargs=schedule)


requeue_failed_tasks = RequeueFailedTasks()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-01-13 11:13
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("djcelery", "0001_initial"),
        ("scheduler", "0002_auto_20161012_2010"),
    ]

    operations = [
        migrations.CreateModel(
            name="QueueTaskRun",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("task_id", models.UUIDField()),
                ("started_at", models.DateTimeField()),
                ("completed_at", models.DateTimeField(null=True)),
                (
                    "celery_cron_definition",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djcelery.CrontabSchedule",
                    ),
                ),
                (
                    "celery_interval_definition",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djcelery.IntervalSchedule",
                    ),
                ),
            ],
        )
    ]



"""
Django settings for seed_scheduler project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

import mimetypes
import os

import dj_database_url
import djcelery
from kombu import Exchange, Queue

# Support SVG on admin
mimetypes.add_type("image/svg+xml", ".svg", True)
mimetypes.add_type("image/svg+xml", ".svgz", True)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get("SECRET_KEY", "REPLACEME")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get("DEBUG", "false").lower() == "true"

ALLOWED_HOSTS = ["*"]


# Application definition

INSTALLED_APPS = (
    # admin
    "django.contrib.admin",
    # core
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # 3rd party
    "raven.contrib.django.raven_compat",
    "rest_framework",
    "rest_framework.authtoken",
    "django_filters",
    "rest_hooks",
    "djcelery",
    "django_prometheus",
    # us
    "scheduler",
)

MIDDLEWARE = (
    "django_prometheus.middleware.PrometheusBeforeMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_prometheus.middleware.PrometheusAfterMiddleware",
)

ROOT_URLCONF = "seed_scheduler.urls"

WSGI_APPLICATION = "seed_scheduler.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    "default": dj_database_url.config(
        default=os.environ.get(
            "SCHEDULER_DATABASE", "postgres://postgres:@localhost/seed_scheduler"
        ),
        engine="django_prometheus.db.backends.postgresql",
    )
}

PROMETHEUS_EXPORT_MIGRATIONS = False


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = "en-gb"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "django.contrib.staticfiles.finders.FileSystemFinder",
)

STATIC_ROOT = "staticfiles"
STATIC_URL = "/static/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

# Sentry configuration
RAVEN_CONFIG = {
    # DevOps will supply you with this.
    "dsn": os.environ.get("SCHEDULER_SENTRY_DSN", None)
}

# REST Framework conf defaults
REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.CursorPagination",
    "PAGE_SIZE": 1000,
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.BasicAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
}

# Webhook event definition
HOOK_EVENTS = {
    # 'any.event.name': 'App.Model.Action' (created/updated/deleted)
    "schedule.added": "scheduler.Schedule.created+"
}

HOOK_DELIVERER = "scheduler.tasks.deliver_hook_wrapper"

HOOK_AUTH_TOKEN = os.environ.get("HOOK_AUTH_TOKEN", "REPLACEME")

# Celery configuration options
CELERY_RESULT_BACKEND = "djcelery.backends.database:DatabaseBackend"
CELERYBEAT_SCHEDULER = "djcelery.schedulers.DatabaseScheduler"

BROKER_URL = os.environ.get("BROKER_URL", "redis://localhost:6379/0")

CELERY_DEFAULT_QUEUE = "seed_scheduler"
CELERY_QUEUES = (
    Queue("seed_scheduler", Exchange("seed_scheduler"), routing_key="seed_scheduler"),
)

CELERY_ALWAYS_EAGER = False

# Tell Celery where to find the tasks
CELERY_IMPORTS = ("scheduler.tasks",)

CELERY_CREATE_MISSING_QUEUES = True
CELERY_ROUTES = {
    "celery.backend_cleanup": {"queue": "mediumpriority"},
    "scheduler.tasks.DeliverHook": {"queue": "priority"},
    "seed_scheduler.scheduler.tasks.queue_tasks": {"queue": "priority"},
    "seed_scheduler.scheduler.tasks.requeue_failed_tasks": {"queue": "priority"},
    "seed_scheduler.scheduler.tasks.deliver_task": {"queue": "lowpriority"},
    "seed_scheduler.scheduler.tasks.fire_metric": {"queue": "metrics"},
}

METRICS_REALTIME = [
    "schedules.created.sum",
    "scheduler.deliver_task.connection_error.sum",
    "scheduler.deliver_task.http_error.400.sum",
    "scheduler.deliver_task.http_error.401.sum",
    "scheduler.deliver_task.http_error.403.sum",
    "scheduler.deliver_task.http_error.404.sum",
    "scheduler.deliver_task.http_error.500.sum",
    "scheduler.deliver_task.timeout.sum",
]
METRICS_SCHEDULED = []
METRICS_SCHEDULED_TASKS = []

CELERY_TASK_SERIALIZER = "json"
CELERY_RESULT_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = ["json"]
CELERY_IGNORE_RESULT = True
CELERYD_MAX_TASKS_PER_CHILD = 50

djcelery.setup_loader()

METRICS_URL = os.environ.get("METRICS_URL", None)
METRICS_AUTH = (
    os.environ.get("METRICS_AUTH_USER", "REPLACEME"),
    os.environ.get("METRICS_AUTH_PASSWORD", "REPLACEME"),
)

DEFAULT_REQUEST_TIMEOUT = float(os.environ.get("DEFAULT_REQUEST_TIMEOUT", 30))
DEFAULT_CLOCK_SKEW_SECONDS = int(os.environ.get("DEFAULT_CLOCK_SKEW_SECONDS", 5))



import json
from datetime import timedelta
from uuid import uuid4

import responses
from django.contrib.auth.models import Group, User
from django.core.management import call_command
from django.test import TestCase, override_settings
from django.utils import timezone
from django.utils.six import StringIO
from djcelery.models import CrontabSchedule, IntervalSchedule, PeriodicTask
from freezegun import freeze_time
from requests_testadapter import TestAdapter, TestSession
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient
from rest_hooks.models import Hook

from seed_scheduler import celery_app

from .models import QueueTaskRun, Schedule, ScheduleFailure
from .tasks import deliver_task, fire_metric, queue_tasks, requeue_failed_tasks

try:
    from urllib.parse import urlparse, urlencode
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode


class RecordingAdapter(TestAdapter):

    """ Record the request that was handled by the adapter.
    """

    request = None

    def send(self, request, *args, **kw):
        self.request = request
        return super(RecordingAdapter, self).send(request, *args, **kw)


class APITestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        self.adminclient = APIClient()
        self.session = TestSession()


class AuthenticatedAPITestCase(APITestCase):
    def make_schedule(self):
        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com",
            "payload": {},
        }
        return Schedule.objects.create(**schedule_data)

    def setUp(self):
        super(AuthenticatedAPITestCase, self).setUp()

        self.username = "testuser"
        self.password = "testpass"
        self.user = User.objects.create_user(
            self.username, "testuser@example.com", self.password
        )
        token = Token.objects.create(user=self.user)
        self.token = token.key
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)
        self.superuser = User.objects.create_superuser(
            "testsu", "su@example.com", "dummypwd"
        )
        sutoken = Token.objects.create(user=self.superuser)
        self.adminclient.credentials(HTTP_AUTHORIZATION="Token %s" % sutoken)


class TestSchedulerAppAPI(AuthenticatedAPITestCase):
    def test_login(self):
        request = self.client.post(
            "/api/token-auth/", {"username": "testuser", "password": "testpass"}
        )
        token = request.data.get("token", None)
        self.assertIsNotNone(
            token, "Could not receive authentication token on login post."
        )
        self.assertEqual(
            request.status_code,
            200,
            "Status code on /api/token-auth was %s -should be 200"
            % request.status_code,
        )

    def test_user_list_pagination(self):
        User.objects.create_user("testuser2", "testuser2@example.com", self.password)

        response = self.client.get("/api/v1/user/")
        self.assertEqual(response.status_code, 200)

        # Test first page
        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["username"], "testuser2")
        self.assertEqual(body["results"][1]["username"], "testsu")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Test next page
        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["username"], "testuser")
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Test previous page
        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["username"], "testuser2")
        self.assertEqual(body["results"][1]["username"], "testsu")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    def test_group_list_pagination(self):
        for i in range(1, 4):
            Group.objects.create(name="group_%s" % i)

        response = self.client.get("/api/v1/group/")
        self.assertEqual(response.status_code, 200)

        # Test first page
        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["name"], "group_3")
        self.assertEqual(body["results"][1]["name"], "group_2")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Test next page
        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["name"], "group_1")
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Test previous page
        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["name"], "group_3")
        self.assertEqual(body["results"][1]["name"], "group_2")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    def test_schedule_list_pagination_one_page(self):
        schedule = self.make_schedule()

        response = self.client.get("/api/v1/schedule/")

        body = response.json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["id"], str(schedule.id))
        self.assertIsNone(body["previous"])
        self.assertIsNone(body["next"])

    def test_schedule_list_pagination_two_pages(self):
        schedules = []
        for i in range(3):
            schedules.append(self.make_schedule())

        # Test first page
        response = self.client.get("/api/v1/schedule/")

        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], str(schedules[2].id))
        self.assertEqual(body["results"][1]["id"], str(schedules[1].id))
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Test next page
        response = self.client.get(body["next"])

        body = response.json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["id"], str(schedules[0].id))
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Test going back to previous page works
        response = self.client.get(body["previous"])

        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], str(schedules[2].id))
        self.assertEqual(body["results"][1]["id"], str(schedules[1].id))
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    def test_create_schedule_cron(self):
        post_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com",
            "payload": {},
            "auth_token": "blah",
        }
        response = self.client.post(
            "/api/v1/schedule/", json.dumps(post_data), content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Schedule.objects.last()
        self.assertEqual(d.frequency, 2)
        self.assertEqual(d.cron_definition, "25 * * * *")
        self.assertIsNotNone(d.celery_cron_definition)
        self.assertEqual(d.celery_cron_definition.minute, "25")
        self.assertEqual(d.auth_token, "blah")

    def test_create_schedule_cron_mappings(self):
        base_schedule = {
            "frequency": 2,
            "cron_definition": "",
            "interval_definition": None,
            "endpoint": "http://example.com",
            "payload": {},
        }
        schedule1 = base_schedule.copy()
        schedule1["cron_definition"] = "25 * * * * *"
        d = Schedule.objects.create(**schedule1)
        self.assertEqual(d.celery_cron_definition.minute, "25")

        schedule2 = base_schedule.copy()
        schedule2["cron_definition"] = "* 14 * * * *"
        d = Schedule.objects.create(**schedule2)
        self.assertEqual(d.celery_cron_definition.hour, "14")

        schedule3 = base_schedule.copy()
        schedule3["cron_definition"] = "* * * * 1 *"
        d = Schedule.objects.create(**schedule3)
        self.assertEqual(d.celery_cron_definition.day_of_week, "1")

        schedule4 = base_schedule.copy()
        schedule4["cron_definition"] = "* * 14 * * *"
        d = Schedule.objects.create(**schedule4)
        self.assertEqual(d.celery_cron_definition.day_of_month, "14")

        schedule5 = base_schedule.copy()
        schedule5["cron_definition"] = "* * * 10 * *"
        d = Schedule.objects.create(**schedule5)
        self.assertEqual(d.celery_cron_definition.month_of_year, "10")

    def test_create_schedule_cron_failed(self):
        post_data = {
            "frequency": 2,
            "cron_definition": "99 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com",
            "payload": {},
        }
        response = self.client.post(
            "/api/v1/schedule/", json.dumps(post_data), content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "cron_definition": [
                    "99 * * * * is not a valid crontab string: item value 99 "
                    "out of range [0, 59]"
                ]
            },
        )

    def test_create_schedule_interval(self):
        post_data = {
            "frequency": 2,
            "cron_definition": None,
            "interval_definition": "1 minutes",
            "endpoint": "http://example.com",
            "payload": {},
        }
        response = self.client.post(
            "/api/v1/schedule/", json.dumps(post_data), content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Schedule.objects.last()
        self.assertEqual(d.frequency, 2)
        self.assertEqual(d.interval_definition, "1 minutes")
        self.assertIsNotNone(d.celery_interval_definition)
        self.assertEqual(d.celery_interval_definition.every, 1)
        self.assertEqual(d.celery_interval_definition.period, "minutes")

    def test_create_schedule_interval_failed(self):
        post_data = {
            "frequency": 2,
            "cron_definition": None,
            "interval_definition": "every one mins",
            "endpoint": "http://example.com",
            "payload": {},
        }
        response = self.client.post(
            "/api/v1/schedule/", json.dumps(post_data), content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {
                "interval_definition": [
                    "every one mins is not a valid interval string: "
                    "integer and period (from: days, hours, minutes, "
                    "seconds, microseconds) e.g. 1 minutes"
                ]
            },
        )

    def test_update_schedule(self):
        # Setup
        s = self.make_schedule()
        post_data = {"enabled": False}
        # Precheck
        self.assertEqual(s.enabled, True)
        self.assertEqual(s.frequency, 2)
        self.assertEqual(s.cron_definition, "25 * * * *")
        self.assertEqual(s.interval_definition, None)
        self.assertEqual(s.endpoint, "http://example.com")
        # Execute
        response = self.client.patch(
            "/api/v1/schedule/%s/" % s.id,
            json.dumps(post_data),
            content_type="application/json",
        )
        s.refresh_from_db()
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(s.enabled, False)
        self.assertEqual(s.frequency, 2)
        self.assertEqual(s.cron_definition, "25 * * * *")
        self.assertEqual(s.interval_definition, None)
        self.assertEqual(s.endpoint, "http://example.com")

    def test_create_webhook(self):
        # Setup
        user = User.objects.get(username="testuser")
        post_data = {
            "target": "http://example.com/registration/",
            "event": "schedule.added",
        }
        # Execute
        response = self.client.post(
            "/api/v1/webhook/", json.dumps(post_data), content_type="application/json"
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = Hook.objects.last()
        self.assertEqual(d.target, "http://example.com/registration/")
        self.assertEqual(d.user, user)

    # This test is not working despite the code working fine
    # If you run these same steps below interactively the webhook will fire
    # @responses.activate
    # def test_webhook(self):
    #     # Setup
    #     post_save.connect(receiver=model_saved, sender=DummyModel,
    #                       dispatch_uid='instance-saved-hook')
    #     Hook.objects.create(user=self.adminuser,
    #                         event='dummymodel.added',
    #                         target='http://example.com/registration/')
    #
    #     expected_webhook = {
    #         "hook": {
    #             "target": "http://example.com/registration/",
    #             "event": "dummymodel.added",
    #             "id": 3
    #         },
    #         "data": {
    #         }
    #     }
    #     responses.add(
    #         responses.POST,
    #         "http://example.com/registration/",
    #         json.dumps(expected_webhook),
    #         status=200, content_type='application/json')
    #     dummymodel_data = {
    #         "product_code": "BLAHBLAH",
    #         "data": {"stuff": "nonsense"}
    #     }
    #     dummy = DummyModel.objects.create(**dummymodel_data)
    #     # Execute
    #     self.assertEqual(responses.calls[0].request.url,
    #                      "http://example.com/registration/")


class TestSchedudlerTasks(AuthenticatedAPITestCase):
    @responses.activate
    def test_deliver_task(self):
        # Tests the deliver task directly
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # Execute
        result = deliver_task.apply_async(
            kwargs={
                "schedule_id": str(schedule.id),
                "auth_token": schedule.auth_token,
                "endpoint": schedule.endpoint,
                "payload": schedule.payload,
            }
        )

        # Check
        self.assertEqual(result.get(), True)
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")

    @responses.activate
    def test_queue_tasks_one_crontab(self):
        # Tests crontab based task runs
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")

    @responses.activate
    def test_queue_tasks_one_interval(self):
        # Tests interval based task runs
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": None,
            "interval_definition": "1 minutes",
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "interval",
                "lookup_id": schedule.celery_interval_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")

    @responses.activate
    def test_queue_tasks_one_not_enabled(self):
        # Tests that with two schedules, one disabled it just runs active
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        Schedule.objects.create(**schedule_data)
        schedule_data = {
            "frequency": 10,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/notrun/",
            "payload": {"run": 1},
        }
        donotrun = Schedule.objects.create(**schedule_data)
        donotrun.enabled = False
        donotrun.save()

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": donotrun.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")

    @responses.activate
    def test_queue_tasks_one_with_none_frequency(self):
        # Tests that with two schedules, one with a None frequency, it runs two
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )
        responses.add(
            responses.POST,
            "http://example.com/runnone/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        Schedule.objects.create(**schedule_data)
        schedule_data = {
            "frequency": None,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/runnone/",
            "payload": {"run": 1},
        }
        runnone = Schedule.objects.create(**schedule_data)
        runnone.save()

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": runnone.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <2> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/runnone/")
        self.assertEqual(responses.calls[1].request.url, "http://example.com/trigger/")

    @responses.activate
    def test_queue_tasks_repeat(self):
        # Tests crontab based task runs
        # Setup
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")

        # Try run again
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertIn("Aborted Queuing", result.get())
        self.assertEqual(QueueTaskRun.objects.all().count(), 1)

    @responses.activate
    def test_queue_tasks_repeat_after_interval(self):
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # Create a previous run
        d1 = timezone.now()
        d1 = d1.replace(minute=20)
        QueueTaskRun.objects.create(
            celery_cron_definition=schedule.celery_cron_definition,
            task_id=uuid4(),
            started_at=d1 - timedelta(hours=1),
            completed_at=d1 + timedelta(minutes=5),
        )

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")
        self.assertEqual(QueueTaskRun.objects.all().count(), 2)

    @responses.activate
    @freeze_time("2017-01-01 17:24:00")
    @override_settings(DEFAULT_CLOCK_SKEW_SECONDS=120)  # 2 minutes
    def test_queue_tasks_repeat_clock_skew(self):
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # We allow for 2 minutes of clock skew in DEFAULT_CLOCK_SKEW_SECONDS
        # With current date the next run should only be in 1 minute, this is
        # inside our 2 minute allowance and we let it run

        # Create a previous run
        d1 = timezone.now()
        QueueTaskRun.objects.create(
            celery_cron_definition=schedule.celery_cron_definition,
            task_id=uuid4(),
            started_at=d1 - timedelta(minutes=59),
            completed_at=d1,
        )

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertEqual(result.get(), "Queued <1> Tasks")
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")
        self.assertEqual(QueueTaskRun.objects.all().count(), 2)

    @responses.activate
    @freeze_time("2017-01-01 17:24:00")
    @override_settings(DEFAULT_CLOCK_SKEW_SECONDS=10)  # 10 seconds
    def test_queue_tasks_repeat_clock_skew_stop(self):
        expected_body = {"run": 1}
        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "frequency": 2,
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)

        # We allow for 10 seconds of clock skew in DEFAULT_CLOCK_SKEW_SECONDS
        # With current date the next run should only be in 1 minute, this is
        # not inside our 10 second allowance and we it is stopped

        # Create a previous run
        d1 = timezone.now()
        QueueTaskRun.objects.create(
            celery_cron_definition=schedule.celery_cron_definition,
            task_id=uuid4(),
            started_at=d1 - timedelta(minutes=59),
            completed_at=d1,
        )

        # Execute
        result = queue_tasks.apply_async(
            kwargs={
                "schedule_type": "crontab",
                "lookup_id": schedule.celery_cron_definition.id,
            }
        )

        # Check
        self.assertIn("Aborted Queuing", result.get())
        self.assertEqual(QueueTaskRun.objects.all().count(), 1)

    @responses.activate
    def test_requeue_failed_tasks(self):
        expected_body = {"run": 1}

        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)
        ScheduleFailure.objects.create(
            schedule=schedule,
            task_id=uuid4(),
            initiated_at=timezone.now(),
            reason="Error",
        )

        requeue_failed_tasks()

        # Check
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")
        self.assertEqual(ScheduleFailure.objects.all().count(), 0)


class TestMetricsAPI(AuthenticatedAPITestCase):
    def test_metrics_read(self):
        # Setup
        # Execute
        response = self.client.get("/api/metrics/", content_type="application/json")
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data["metrics_available"],
            [
                "schedules.created.sum",
                "scheduler.deliver_task.connection_error.sum",
                "scheduler.deliver_task.http_error.400.sum",
                "scheduler.deliver_task.http_error.401.sum",
                "scheduler.deliver_task.http_error.403.sum",
                "scheduler.deliver_task.http_error.404.sum",
                "scheduler.deliver_task.http_error.500.sum",
                "scheduler.deliver_task.timeout.sum",
            ],
        )

    @responses.activate
    def test_post_metrics(self):
        # Setup
        # deactivate Testsession for this test
        self.session = None
        responses.add(
            responses.POST,
            "http://metrics-url/metrics/",
            json={"foo": "bar"},
            status=200,
            content_type="application/json",
        )
        # Execute
        response = self.client.post("/api/metrics/", content_type="application/json")
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["scheduled_metrics_initiated"], True)


class TestMetrics(AuthenticatedAPITestCase):
    def check_request(self, request, method, params=None, data=None, headers=None):
        self.assertEqual(request.method, method)
        if params is not None:
            url = urlparse.urlparse(request.url)
            qs = urlparse.parse_qsl(url.query)
            self.assertEqual(dict(qs), params)
        if headers is not None:
            for key, value in headers.items():
                self.assertEqual(request.headers[key], value)
        if data is None:
            self.assertEqual(request.body, None)
        else:
            self.assertEqual(json.loads(request.body), data)

    def add_metrics_response(self):
        """
        Adds a response for any requests to the metrics API endpoint.
        """
        responses.add(
            responses.POST, "http://metrics-url/metrics/", status=201, json={}
        )

    @responses.activate
    def test_direct_fire(self):
        # Setup
        self.add_metrics_response()
        # Execute
        result = fire_metric.apply_async(
            kwargs={"metric_name": "foo.last", "metric_value": 1}
        )
        # Check
        self.check_request(responses.calls[0].request, "POST", data={"foo.last": 1.0})
        self.assertEqual(result.get(), "Fired metric <foo.last> with value <1.0>")


class TestUserCreation(AuthenticatedAPITestCase):
    def test_create_user_and_token(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post("/api/v1/user/token/", user_request)
        token = request.json().get("token", None)
        # Check
        self.assertIsNotNone(token, "Could not receive authentication token on post.")
        self.assertEqual(
            request.status_code,
            201,
            "Status code on /api/v1/user/token/ was %s (should be 201)."
            % request.status_code,
        )

    def test_create_user_and_token_fail_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.client.post("/api/v1/user/token/", user_request)
        error = request.json().get("detail", None)
        # Check
        self.assertIsNotNone(error, "Could not receive error on post.")
        self.assertEqual(
            error,
            "You do not have permission to perform this action.",
            "Error message was unexpected: %s." % error,
        )

    def test_create_user_and_token_not_created(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post("/api/v1/user/token/", user_request)
        token = request.json().get("token", None)
        # And again, to get the same token
        request2 = self.adminclient.post("/api/v1/user/token/", user_request)
        token2 = request2.json().get("token", None)

        # Check
        self.assertEqual(
            token, token2, "Tokens are not equal, should be the same as not recreated."
        )

    def test_create_user_new_token_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        request = self.adminclient.post("/api/v1/user/token/", user_request)
        token = request.json().get("token", None)
        cleanclient = APIClient()
        cleanclient.credentials(HTTP_AUTHORIZATION="Token %s" % token)
        # Execute
        request = cleanclient.post("/api/v1/user/token/", user_request)
        error = request.json().get("detail", None)
        # Check
        # new user should not be admin
        self.assertIsNotNone(error, "Could not receive error on post.")
        self.assertEqual(
            error,
            "You do not have permission to perform this action.",
            "Error message was unexpected: %s." % error,
        )


class TestHealthcheckAPI(AuthenticatedAPITestCase):
    def test_healthcheck_read(self):
        # Setup
        # Execute
        response = self.client.get("/api/health/", content_type="application/json")
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["up"], True)
        self.assertEqual(response.data["result"]["database"], "Accessible")


@celery_app.task
def noop(argument):
    return argument


class TestTriggerPeriodicTask(TestCase):

    timeout = 1

    def test_trigger_periodic_task(self):
        cs = CrontabSchedule.objects.create(
            **{
                "minute": "*",
                "hour": "*",
                "day_of_week": "*",
                "day_of_month": "*",
                "month_of_year": "*",
            }
        )

        pt = PeriodicTask.objects.create(
            **{
                "name": "noop task",
                "task": "scheduler.tests.noop",
                "crontab": cs,
                "enabled": True,
                "args": '["hello world"]',
            }
        )
        stdout = StringIO()
        stderr = StringIO()
        self.assertFalse(pt.last_run_at)
        call_command(
            "trigger_periodic_task",
            str(pt.pk),
            "--confirm",
            stdout=stdout,
            stderr=stderr,
        )
        self.assertEqual(stdout.getvalue().strip(), "hello world")
        pt_after_run = PeriodicTask.objects.get(pk=pt.pk)
        self.assertTrue(pt_after_run.last_run_at)


class TestTriggerDeliverTasks(TestCase):

    timeout = 1

    def setUp(self):
        self.session = TestSession()
        self.crontab_schedule = CrontabSchedule.objects.create(
            **{
                "minute": "*",
                "hour": "*",
                "day_of_week": "*",
                "day_of_month": "*",
                "month_of_year": "*",
            }
        )

        self.periodic_crontab_task = PeriodicTask.objects.create(
            **{
                "name": "noop crontab task",
                "task": "scheduler.tests.noop",
                "crontab": self.crontab_schedule,
                "enabled": True,
                "args": '["hello world"]',
            }
        )

        self.interval_schedule = IntervalSchedule.objects.create(
            **{"every": 5, "period": "minutes"}
        )

        self.periodic_interval_task = PeriodicTask.objects.create(
            **{
                "name": "noop interval task",
                "task": "scheduler.tests.noop",
                "interval": self.interval_schedule,
                "enabled": True,
                "args": '["hello world"]',
            }
        )

    def mount_subscription(self, subscription_id, msisdns, identity_id=None):
        identity_id = identity_id or ("identity-for-%s" % (subscription_id,))
        responses.add(
            responses.GET,
            "http://sbm.example.org/api/v1/subscriptions/%s/" % (subscription_id,),
            status=200,
            json={"identity": identity_id},
        )

        addresses = [(msisdns[0], {"default": True})]
        for msisdn in msisdns[1:]:
            addresses.append((msisdn, {}))

        responses.add(
            responses.GET,
            "http://idstore.example.org/api/v1/identities/%s/" % (identity_id,),
            status=200,
            json={
                "details": {
                    "addresses": {"msisdn": dict(addresses)},
                    "default_addr_type": "msisdn",
                }
            },
        )

    def mount_outbound(self, msisdn, since, until, outbound_count=0):
        responses.add(
            responses.GET,
            "http://sender.example.org/api/v1/outbound/?%s"
            % (
                urlencode(
                    {
                        "to_addr": msisdn,
                        "after": since.isoformat(),
                        "before": until.isoformat(),
                    }
                ),
            ),
            status=200,
            body=json.dumps(
                {
                    "results": [
                        {"to_addr": msisdn, "content": counter}
                        for counter in range(outbound_count)
                    ]
                }
            ),
            match_querystring=True,
        )

    def mount_sbm_send(self, subscription_id):
        responses.add(
            responses.POST,
            "http://sbm.example.org/api/v1/subscriptions/%s/send" % (subscription_id,),
            status=200,
            body="{}",
        )

    def sbm_send_happened(self, subscription_id):
        sbm_url = "http://sbm.example.org/api/v1/subscriptions/%s/send" % (
            subscription_id,
        )
        for call in responses.calls:
            request = call.request
            if request.url == sbm_url:
                return True
        return False

    def assertSBMSend(self, subscription_id):
        self.assertTrue(
            self.sbm_send_happened(subscription_id),
            "Expected SBM to send for %s but didn't." % (subscription_id,),
        )

    def assertNotSBMSend(self, subscription_id):
        self.assertFalse(
            self.sbm_send_happened(subscription_id),
            "Expected SBM to not send for %s but did." % (subscription_id,),
        )

    @freeze_time("2017-01-01 17:24:00")
    @responses.activate
    def test_deliver_tasks_interval(self):

        since = timezone.now() - timedelta(days=1)
        until = timezone.now() + timedelta(days=1)

        stdout = StringIO()
        stderr = StringIO()

        Schedule.objects.create(
            enabled=True,
            celery_interval_definition=self.interval_schedule,
            endpoint=(
                "http://sbm.example.org/api/v1/subscriptions"
                "/subscription-uuid-1/send"
            ),
            auth_token="sbm_token",
            payload={},
        )

        self.mount_subscription("subscription-uuid-1", ["+27000000000", "+27111111111"])
        self.mount_outbound("+27000000000", since=since, until=until, outbound_count=1)
        self.mount_outbound("+27111111111", since=since, until=until, outbound_count=0)

        resend = Schedule.objects.create(
            enabled=True,
            celery_interval_definition=self.interval_schedule,
            endpoint=(
                "http://sbm.example.org/api/v1/subscriptions"
                "/subscription-uuid-2/send"
            ),
            auth_token="sbm_token",
            payload={},
        )

        self.mount_subscription("subscription-uuid-2", ["+27222222222"])
        self.mount_outbound("+27222222222", since=since, until=until, outbound_count=0)
        self.mount_sbm_send("subscription-uuid-2")

        self.assertIsNone(resend.last_run)

        call_command(
            "trigger_deliver_tasks",
            "--identity-store-token",
            "token",
            "--identity-store-url",
            "http://idstore.example.org/api/v1",
            "--message-sender-token",
            "token",
            "--message-sender-url",
            "http://sender.example.org/api/v1",
            "--since",
            since.isoformat(),
            "--until",
            until.isoformat(),
            "--confirm",
            "interval:%s" % (self.interval_schedule.pk,),
            stdout=stdout,
            stderr=stderr,
        )

        self.assertNotSBMSend("subscription-uuid-1")
        self.assertSBMSend("subscription-uuid-2")
        self.assertEqual(stdout.getvalue().strip(), str(resend))

        resend.refresh_from_db()
        self.assertEqual(resend.last_run, timezone.now())

    @responses.activate
    def test_deliver_tasks_dry_run(self):

        since = timezone.now() - timedelta(days=1)
        until = timezone.now() + timedelta(days=1)

        stdout = StringIO()
        stderr = StringIO()

        resend = Schedule.objects.create(
            enabled=True,
            celery_interval_definition=self.interval_schedule,
            endpoint=(
                "http://sbm.example.org/api/v1/subscriptions" "/subscription-uuid/send"
            ),
            auth_token="sbm_token",
            payload={},
        )

        self.mount_subscription("subscription-uuid", ["+27222222222"])
        self.mount_outbound("+27222222222", since=since, until=until, outbound_count=0)

        call_command(
            "trigger_deliver_tasks",
            "--identity-store-token",
            "token",
            "--identity-store-url",
            "http://idstore.example.org/api/v1",
            "--message-sender-token",
            "token",
            "--message-sender-url",
            "http://sender.example.org/api/v1",
            "--since",
            since.isoformat(),
            "--until",
            until.isoformat(),
            "--confirm",
            "--dry-run",
            "interval:%s" % (self.interval_schedule.pk,),
            stdout=stdout,
            stderr=stderr,
        )

        self.assertNotSBMSend("subscription-uuid")
        self.assertEqual(stdout.getvalue().strip(), "Dry run for %s" % (str(resend),))


class TestFailedTaskAPI(AuthenticatedAPITestCase):
    def test_list_failed_tasks(self):
        schedule_data = {
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)
        failures = []
        for i in range(3):
            failures.append(
                ScheduleFailure.objects.create(
                    schedule=schedule,
                    task_id=uuid4(),
                    initiated_at=timezone.now(),
                    reason="Error",
                )
            )

        response = self.client.get(
            "/api/v1/failed-tasks/", content_type="application/json"
        )
        self.assertEqual(response.status_code, 200)

        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], failures[2].id)
        self.assertEqual(body["results"][1]["id"], failures[1].id)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["id"], failures[0].id)
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], failures[2].id)
        self.assertEqual(body["results"][1]["id"], failures[1].id)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    @responses.activate
    def test_failed_tasks_requeue(self):
        expected_body = {"run": 1}

        responses.add(
            responses.POST,
            "http://example.com/trigger/",
            json.dumps(expected_body),
            status=200,
            content_type="application/json",
        )

        schedule_data = {
            "cron_definition": "25 * * * *",
            "interval_definition": None,
            "endpoint": "http://example.com/trigger/",
            "payload": {"run": 1},
        }
        schedule = Schedule.objects.create(**schedule_data)
        ScheduleFailure.objects.create(
            schedule=schedule,
            task_id=uuid4(),
            initiated_at=timezone.now(),
            reason="Error",
        )

        response = self.client.post(
            "/api/v1/failed-tasks/", content_type="application/json"
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["requeued_failed_tasks"], True)
        self.assertEqual(responses.calls[0].request.url, "http://example.com/trigger/")
        self.assertEqual(ScheduleFailure.objects.all().count(), 0)



from django.contrib.auth.models import Group, User
from rest_framework import mixins, status, viewsets
from rest_framework.authtoken.models import Token
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import IsAdminUser, IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_hooks.models import Hook

from seed_scheduler.utils import get_available_metrics

from .models import Schedule, ScheduleFailure
from .serializers import (
    CreateUserSerializer,
    GroupSerializer,
    HookSerializer,
    ScheduleFailureSerializer,
    ScheduleSerializer,
    UserSerializer,
)
from .tasks import requeue_failed_tasks

# Uncomment line below if scheduled metrics are added
# from .tasks import scheduled_metrics


class CreatedAtCursorPagination(CursorPagination):
    ordering = "-created_at"


class IdCursorPagination(CursorPagination):
    ordering = "-id"


class HookViewSet(viewsets.ModelViewSet):
    """
    Retrieve, create, update or destroy webhooks.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Hook.objects.all()
    serializer_class = HookSerializer
    pagination_class = IdCursorPagination

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class UserViewSet(viewsets.ReadOnlyModelViewSet):

    """
    API endpoint that allows users to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = User.objects.all()
    serializer_class = UserSerializer
    pagination_class = IdCursorPagination


class GroupViewSet(viewsets.ReadOnlyModelViewSet):

    """
    API endpoint that allows groups to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Group.objects.all()
    serializer_class = GroupSerializer
    pagination_class = IdCursorPagination


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """

    permission_classes = (IsAdminUser,)

    def post(self, request):
        """Create a user and token, given an email. If user exists just
        provide the token."""
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get("email")
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(status=status.HTTP_201_CREATED, data={"token": token.key})


class ScheduleViewSet(viewsets.ModelViewSet):

    """
    API endpoint that allows schedule models to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Schedule.objects.all()
    serializer_class = ScheduleSerializer
    pagination_class = CreatedAtCursorPagination

    def perform_create(self, serializer):
        serializer.save(created_by=self.request.user, updated_by=self.request.user)

    def perform_update(self, serializer):
        serializer.save(updated_by=self.request.user)


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {"metrics_available": get_available_metrics()}
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        # Uncomment line below if scheduled metrics are added
        # scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_scheduler
        import django
        import rest_framework

        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_scheduler.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__,
                },
            },
        }
        return Response(resp, status=status)


class FailedTaskViewSet(
    mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet
):
    permission_classes = (IsAuthenticated,)
    queryset = ScheduleFailure.objects.all()
    serializer_class = ScheduleFailureSerializer
    pagination_class = IdCursorPagination

    def create(self, request):
        status = 201
        resp = {"requeued_failed_tasks": True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)



import os
import re

from django.core.management import BaseCommand, CommandError
from django.core.validators import URLValidator
from django.utils.dateparse import parse_datetime
from django.utils.six.moves import input

from scheduler.models import Schedule
from scheduler.tasks import DeliverTask

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


def get_schedule(schedule_string):
    schedule_type, _, lookup_id = schedule_string.partition(":")
    if schedule_type == "crontab":
        kwargs = {"celery_cron_definition": int(lookup_id)}
    elif schedule_type == "interval":
        kwargs = {"celery_interval_definition": int(lookup_id)}
    else:
        raise CommandError(
            "Invalid schedule format: %s, unknown type: %s"
            % (schedule_string, schedule_type)
        )
    return Schedule.objects.filter(enabled=True).filter(**kwargs)


def get_model(model_class):
    def callback(pk):
        try:
            return model_class.objects.get(pk=int(pk))
        except (model_class.DoesNotExist,) as e:
            raise CommandError(e)
        except (Exception,) as e:
            raise CommandError(e)

    return callback


def mk_validator(validator_class):
    def validator_callback(input_str):
        validator = validator_class()
        validator(input_str)
        return input_str

    return validator_callback


class Command(BaseCommand):
    help = (
        "Manually generate the deliver_tasks for a queue_tasks job "
        "that failed for some reason.\n\n"
        "* Only use this if things went HORRIBLY WRONG *"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--message-sender-token",
            type=str,
            default=os.environ.get("MESSAGE_SENDER_TOKEN"),
            help=("The authorization token for the " "Message Sender"),
        )
        parser.add_argument(
            "--message-sender-url",
            type=mk_validator(URLValidator),
            default=os.environ.get("MESSAGE_SENDER_URL"),
            help=("The URL for the Message Sender API"),
        )
        parser.add_argument(
            "--identity-store-token",
            type=str,
            default=os.environ.get("IDENTITY_STORE_TOKEN"),
            help="The token for the ID Store",
        )
        parser.add_argument(
            "--identity-store-url",
            type=mk_validator(URLValidator),
            default=os.environ.get("IDENTITY_STORE_URL"),
            help="The Identity Store API URL",
        )
        parser.add_argument(
            "--default-addr-type",
            type=str,
            default="msisdn",
            help=(
                "The default address type to assume when not specified. "
                "Defaults to `msisdn`."
            ),
        )
        parser.add_argument(
            "--since",
            type=parse_datetime,
            help=(
                "Filter for outbound created_at since " "(required YYYY-MM-DD HH:MM:SS)"
            ),
        )
        parser.add_argument(
            "--until",
            type=parse_datetime,
            help=("Filter for created_at until " "(required YYYY-MM-DD HH:MM:SS)"),
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Do a dry-run, do not actually dispatch the tasks.",
        )
        parser.add_argument(
            "--confirm",
            action="store_true",
            default=False,
            help="Agree to everything, I know what I am doing.",
        )
        parser.add_argument(
            "schedule",
            type=get_schedule,
            help=(
                "The schedule to run deliver_tasks for. "
                "The format is <definition>:<celery_definition_lookup_id>. "
                "Valid options for definitions are `crontab` or `interval`, "
                "e.g.: `crontab:1` or `interval:2`"
            ),
        )
        try:
            import seed_services_client  # noqa
        except (ImportError,):
            raise CommandError(
                "Please make sure you have the seed-services-client "
                "package installed."
            )

    def handle(self, *args, **options):
        schedules = options["schedule"]
        default_addr_type = options["default_addr_type"]
        identity_store_token = options["identity_store_token"]
        identity_store_url = options["identity_store_url"]
        message_sender_token = options["message_sender_token"]
        message_sender_url = options["message_sender_url"]
        since = options["since"]
        until = options["until"]
        dry_run = options["dry_run"]

        if not since:
            raise CommandError("--since is a required parameter")

        if not until:
            raise CommandError("--until is a required parameter")

        from seed_services_client import (
            StageBasedMessagingApiClient,
            IdentityStoreApiClient,
            MessageSenderApiClient,
        )

        id_store_client = IdentityStoreApiClient(
            identity_store_token, identity_store_url
        )

        message_sender_client = MessageSenderApiClient(
            message_sender_token, message_sender_url
        )

        def confirm(prompt):
            if options["confirm"]:
                return True
            try:
                return input("%s [y/n] > " % (prompt,)).lower() == "y"
            except KeyboardInterrupt:
                raise CommandError("Please confirm the question.")

        msg = (
            "You are about to trigger a stage based messaging send for %s "
            "subscriptions. Are you sure you want to be doing this?"
            % (self.style.NOTICE(str(schedules.count())))
        )

        if not confirm(msg):
            raise CommandError("Please confirm as you need to know what you are doing.")

        for schedule in schedules.iterator():
            sbm_api_url, subscription_uuid = self.parse_sbm_api_url(schedule.endpoint)
            sbm_client = StageBasedMessagingApiClient(schedule.auth_token, sbm_api_url)
            subscription = sbm_client.get_subscription(subscription_uuid)
            identity = id_store_client.get_identity(subscription["identity"])
            addresses = self.get_addresses(identity, default_addr_type)
            any_outbounds = []
            for address in addresses:
                outbounds = message_sender_client.get_outbounds(
                    params={
                        "to_addr": address,
                        "before": until.isoformat(),
                        "after": since.isoformat(),
                    }
                )
                any_outbounds.extend(outbounds["results"])

            if any_outbounds:
                continue

            if dry_run:
                self.stdout.write("Dry run for %s" % (schedule,))
            else:
                self.stdout.write("%s" % (schedule,))
                DeliverTask.apply_async(
                    kwargs={
                        "schedule_id": str(schedule.id),
                        "auth_token": schedule.auth_token,
                        "endpoint": schedule.endpoint,
                        "payload": schedule.payload,
                    }
                )

    def get_addresses(self, identity, default_addr_type):
        details = identity.get("details", {})
        default_addr_type = details.get("default_addr_type") or default_addr_type
        addresses = details.get("addresses", {}).get(default_addr_type, {})
        return addresses.keys()

    def parse_sbm_api_url(self, endpoint):
        parse_result = urlparse(endpoint)
        (subscription_uuid,) = re.match(
            r"/api/v1/subscriptions/([0-9a-z\-]+)/send", parse_result.path
        ).groups()
        return (
            "%s://%s/api/v1/" % (parse_result.scheme, parse_result.netloc),
            subscription_uuid,
        )



from seed_scheduler.settings import *  # noqa: F403

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "TESTSEKRET"

CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
CELERY_ALWAYS_EAGER = True
BROKER_BACKEND = "memory"
CELERY_RESULT_BACKEND = "djcelery.backends.database:DatabaseBackend"

METRICS_URL = "http://metrics-url"
METRICS_AUTH_TOKEN = "REPLACEME"

PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)

# REST Framework conf defaults
REST_FRAMEWORK["PAGE_SIZE"] = 2  # noqa: F405



# Generated by Django 2.1.4 on 2019-05-09 14:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scheduler", "0005_auto_20181219_0934")]

    operations = [
        migrations.AddField(
            model_name="schedule",
            name="last_run",
            field=models.DateTimeField(null=True),
        )
    ]



from django.contrib.auth.models import Group, User
from rest_framework import serializers
from rest_hooks.models import Hook

from .models import Schedule, ScheduleFailure


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = User
        fields = ("url", "username", "email", "groups")


class GroupSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Group
        fields = ("url", "name")


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class ScheduleSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Schedule
        read_only_fields = (
            "created_by",
            "updated_by",
            "celery_cron_definition",
            "celery_interval_definition",
        )
        fields = (
            "url",
            "id",
            "frequency",
            "cron_definition",
            "interval_definition",
            "endpoint",
            "payload",
            "auth_token",
            "next_send_at",
            "enabled",
            "created_at",
            "created_by",
            "updated_at",
            "updated_by",
        )


class HookSerializer(serializers.ModelSerializer):
    class Meta:
        model = Hook
        read_only_fields = ("user",)
        fields = "__all__"


class ScheduleFailureSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = ScheduleFailure
        fields = ("url", "id", "schedule", "task_id", "initiated_at", "reason")



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-10-12 20:10
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import scheduler.models


class Migration(migrations.Migration):

    dependencies = [("scheduler", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="celery_cron_definition",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="djcelery.CrontabSchedule",
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="celery_interval_definition",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="djcelery.IntervalSchedule",
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="schedules_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="cron_definition",
            field=models.CharField(
                blank=True,
                max_length=500,
                null=True,
                validators=[scheduler.models.validate_crontab],
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="interval_definition",
            field=models.CharField(
                blank=True,
                max_length=100,
                null=True,
                validators=[scheduler.models.validate_interval],
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="payload",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True, default={}, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="schedules_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import uuid

from crontab import CronTab
from django.contrib.auth.models import User
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from djcelery.models import CrontabSchedule, IntervalSchedule, PeriodicTask


def validate_crontab(value):
    try:
        CronTab(value)
    except ValueError as e:
        raise ValidationError(
            _("%(value)s is not a valid crontab string: %(reason)s"),
            params={"value": value, "reason": e},
        )


def validate_interval(value):
    try:
        every, period = value.split()
        int(every)
        if period not in ["days", "hours", "minutes", "seconds", "microseconds"]:
            raise ValidationError(
                _(
                    "%(value)s is not a valid period. Accepted: days, hours, "
                    "minutes, seconds, microseconds)"
                ),
                params={"value": value},
            )
    except ValueError:
        raise ValidationError(
            _(
                "%(value)s is not a valid interval string: integer and "
                "period (from: days, hours, minutes, seconds, microseconds) "
                "e.g. 1 minutes"
            ),
            params={"value": value},
        )


@python_2_unicode_compatible
class Schedule(models.Model):

    """
    Base model with all scheduled tasks
    frequency: number of times task should run in total (deprecated)
    cron_definition: cron syntax of schedule (i.e. 'm h d dM MY')
    interval_definition: integer and period
        (from: days, hours, minutes, seconds, microseconds) e.g. 1 minutes
    endpoint: what URL to POST to
    payload: what json encoded payload to include on the POST
    next_send_at: when the task is next expected to run (not guarenteed)
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    # frequency and triggered are deprecated and will eventually be removed
    frequency = models.IntegerField(null=True, blank=True)
    triggered = models.IntegerField(null=False, blank=False, default=0)
    cron_definition = models.CharField(
        max_length=500, null=True, blank=True, validators=[validate_crontab]
    )
    celery_cron_definition = models.ForeignKey(
        CrontabSchedule, on_delete=models.CASCADE, null=True, blank=True
    )
    interval_definition = models.CharField(
        max_length=100, null=True, blank=True, validators=[validate_interval]
    )
    celery_interval_definition = models.ForeignKey(
        IntervalSchedule, on_delete=models.CASCADE, null=True, blank=True
    )
    endpoint = models.CharField(max_length=500, null=False)
    auth_token = models.CharField(max_length=500, null=True, blank=True)
    payload = JSONField(null=True, blank=True, default=dict)
    next_send_at = models.DateTimeField(null=True, blank=True)
    enabled = models.BooleanField(default=True)
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        User,
        related_name="schedules_created",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    updated_by = models.ForeignKey(
        User,
        related_name="schedules_updated",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    user = property(lambda self: self.created_by)
    last_run = models.DateTimeField(null=True)

    def serialize_hook(self, hook):
        # optional, there are serialization defaults
        # we recommend always sending the Hook
        # metadata along for the ride as well
        # not sending auth token
        return {
            "hook": hook.dict(),
            "data": {
                "id": str(self.id),
                "frequency": self.frequency,
                "triggered": self.triggered,
                "cron_definition": self.cron_definition,
                "interval_definition": self.interval_definition,
                "endpoint": self.endpoint,
                "payload": self.payload,
                "next_send_at": self.next_send_at.isoformat(),
                "enabled": self.enabled,
                "created_at": self.created_at.isoformat(),
                "updated_at": self.updated_at.isoformat(),
            },
        }

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@receiver(pre_save, sender=Schedule)
def schedule_saved(sender, instance, **kwargs):
    if (
        instance.cron_definition is not None
        and instance.cron_definition != ""
        and instance.celery_cron_definition is None
    ):
        # CronTab package just used to parse and validate the string nicely.
        entry = CronTab(instance.cron_definition)
        schedule = {
            "minute": entry.matchers.minute.input,
            "hour": entry.matchers.hour.input,
            "day_of_week": entry.matchers.weekday.input,
            "day_of_month": entry.matchers.day.input,
            "month_of_year": entry.matchers.month.input,
        }
        cs, createdcs = CrontabSchedule.objects.get_or_create(**schedule)
        instance.celery_cron_definition = cs
        if createdcs:
            # make the periodic task
            pt = {
                "name": "Run %s" % instance.cron_definition,
                "task": "seed_scheduler.scheduler.tasks.queue_tasks",
                "crontab": cs,
                "enabled": True,
                "args": '["crontab", %s]' % cs.id,
            }
            PeriodicTask.objects.create(**pt)
    if (
        instance.interval_definition is not None
        and instance.interval_definition != ""
        and instance.celery_interval_definition is None
    ):
        every, period = instance.interval_definition.split()
        interval = {"every": int(every), "period": period}
        intsch, createdsch = IntervalSchedule.objects.get_or_create(**interval)
        instance.celery_interval_definition = intsch
        if createdsch:
            # make the periodic task
            pt = {
                "name": "Run %s" % instance.interval_definition,
                "task": "seed_scheduler.scheduler.tasks.queue_tasks",
                "interval": intsch,
                "enabled": True,
                "args": '["interval", %s]' % intsch.id,
            }
            PeriodicTask.objects.create(**pt)


@python_2_unicode_compatible
class QueueTaskRun(models.Model):
    task_id = models.UUIDField()
    celery_cron_definition = models.ForeignKey(
        CrontabSchedule, on_delete=models.CASCADE, null=True, blank=True
    )
    celery_interval_definition = models.ForeignKey(
        IntervalSchedule, on_delete=models.CASCADE, null=True, blank=True
    )
    started_at = models.DateTimeField()
    completed_at = models.DateTimeField(null=True)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class ScheduleFailure(models.Model):
    schedule = models.ForeignKey(Schedule, on_delete=models.CASCADE)
    task_id = models.UUIDField()
    initiated_at = models.DateTimeField()
    reason = models.TextField()

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)



from django.core.management import BaseCommand, CommandError
from django.utils.six.moves import input
from djcelery.models import PeriodicTask
from djcelery.schedulers import ModelEntry


class Command(BaseCommand):
    help = (
        "Fire djcelery PeriodicTask manually. \n\n"
        "* Only use this if things went HORRIBLY WRONG *"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "periodic-task-id",
            type=int,
            help="What is the PK of the PeriodicTask to fire manually",
        )
        parser.add_argument(
            "--confirm",
            action="store_true",
            default=False,
            help=(
                "Do not ask for any kind of confirmation as " "I know what I am doing"
            ),
        )
        parser.add_argument(
            "--ignore-result",
            action="store_true",
            default=False,
            help=("Do not wait for the task to complete and return " "the result"),
        )
        parser.add_argument(
            "--timeout",
            type=int,
            default=60,
            help=(
                "How long to wait in seconds for the result to return, "
                "set to 0 to disable. Defaults to 60 seconds."
            ),
        )

    def handle(self, *args, **options):
        periodic_task_id = options["periodic-task-id"]
        ignore_result = options["ignore_result"]
        if options["timeout"]:
            timeout = options["timeout"]
        else:
            timeout = None

        def confirm(prompt):
            if options["confirm"]:
                return True
            try:
                return input("%s [y/n] > " % (prompt,)).lower() == "y"
            except KeyboardInterrupt:
                raise CommandError("Please confirm the question.")

        try:
            periodic_task = PeriodicTask.objects.get(pk=periodic_task_id)
        except PeriodicTask.DoesNotExist:
            raise CommandError(
                "PeriodicTask with id %s does not exist." % (periodic_task_id)
            )

        if periodic_task.last_run_at:
            msg = (
                "The task %s was last run on %s\n"
                "Are you sure you want to resubmit this periodic task?"
                % (
                    self.style.NOTICE(periodic_task),
                    self.style.NOTICE(periodic_task.last_run_at),
                )
            )
        else:
            msg = (
                "The task %s has never been run before.\n"
                "Are you sure you want to submit this periodic task?"
                % (self.style.NOTICE(periodic_task),)
            )

        if not confirm(msg):
            raise CommandError("Please confirm as you need to know what you are doing.")

        from celery import current_app

        model_entry = ModelEntry(periodic_task)

        app = current_app._get_current_object()
        task = app.tasks.get(model_entry.task)
        if task:
            async_result = task.apply_async(
                model_entry.args, model_entry.kwargs, **model_entry.options
            )
        else:
            async_result = app.send_task(
                model_entry.task,
                model_entry.args,
                model_entry.kwargs,
                **model_entry.options
            )

        periodic_task.last_run_at = app.now()
        periodic_task.save()

        if not ignore_result:
            result = async_result.get(timeout=timeout)
            self.stdout.write(result)



from django.conf.urls import include, url
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r"user", views.UserViewSet)
router.register(r"group", views.GroupViewSet)
router.register(r"schedule", views.ScheduleViewSet)
router.register(r"webhook", views.HookViewSet)
router.register(r"failed-tasks", views.FailedTaskViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url(r"^api/v1/user/token/$", views.UserView.as_view(), name="create-user-token"),
    url(r"^api/v1/", include(router.urls)),
]



# Generated by Django 2.1.4 on 2018-12-19 09:34

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scheduler", "0004_schedulefailure")]

    operations = [
        migrations.AlterField(
            model_name="schedule",
            name="created_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="schedules_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="payload",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                blank=True, default=dict, null=True
            ),
        ),
        migrations.AlterField(
            model_name="schedule",
            name="updated_by",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="schedules_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



import os

from django.conf.urls import include, url
from django.contrib import admin
from django.urls import path
from django_prometheus import exports as django_prometheus
from rest_framework.authtoken.views import obtain_auth_token
from rest_framework.documentation import include_docs_urls

from scheduler import views
from seed_scheduler.decorators import internal_only

admin.site.site_header = os.environ.get("SCHEDULER_TITLE", "Scheduler Admin")


urlpatterns = [
    path("admin/", admin.site.urls),
    url(r"^api/auth/", include("rest_framework.urls", namespace="rest_framework")),
    url(r"^api/token-auth/", obtain_auth_token),
    url(r"^api/metrics/", views.MetricsView.as_view()),
    url(r"^api/health/", views.HealthcheckView.as_view()),
    url(r"^", include("scheduler.urls")),
    path("docs/", include_docs_urls(title=admin.site.site_header)),
    path(
        "metrics", internal_only(django_prometheus.ExportToDjangoView), name="metrics"
    ),
]



from django.contrib import admin

from .models import Schedule


class ScheduleAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "triggered",
        "cron_definition",
        "interval_definition",
        "endpoint",
        "next_send_at",
        "enabled",
        "created_at",
        "updated_at",
    )
    list_filter = (
        "cron_definition",
        "interval_definition",
        "enabled",
        "next_send_at",
        "created_at",
        "updated_at",
    )
    search_fields = ["id", "endpoint"]


admin.site.register(Schedule, ScheduleAdmin)
